from ..uhid import BUS_USB

DS5_EDGE_VENDOR = 0x054C
DS5_EDGE_PRODUCT = 0x0DF2
DS5_EDGE_VERSION = 0
DS5_EDGE_COUNTRY = 0
DS5_EDGE_BUS = BUS_USB
DS5_EDGE_NAME = b"Emulated Sony DS5 Edge Controller"

DS5_EDGE_MIN_REPORT_FREQ = 25
DS5_EDGE_MAX_REPORT_FREQ = 1000
DS5_EDGE_DELTA_TIME = 4096.0


def _prep_def_report():
    d = bytearray(64)
    d[0] = 0x01

    d[62] = 0x80
    d[57] = 0x80
    d[53] = 0x80
    d[48] = 0x80
    d[35] = 0x80
    d[44] = 0x80

    return bytes(d)


DS5_EDGE_REPORT_USB_BASE = _prep_def_report()

DS5_EDGE_DESCRIPTOR = bytes(
    [
        0x05,
        0x01,  # Usage Page (Generic Desktop)        0
        0x09,
        0x05,  # Usage (Game Pad)                    2
        0xA1,
        0x01,  # Collection (Application)            4
        0x85,
        0x01,  #  Report ID (1)                      6
        0x09,
        0x30,  #  Usage (X)                          8
        0x09,
        0x31,  #  Usage (Y)                          10
        0x09,
        0x32,  #  Usage (Z)                          12
        0x09,
        0x35,  #  Usage (Rz)                         14
        0x09,
        0x33,  #  Usage (Rx)                         16
        0x09,
        0x34,  #  Usage (Ry)                         18
        0x15,
        0x00,  #  Logical Minimum (0)                20
        0x26,
        0xFF,
        0x00,  #  Logical Maximum (255)              22
        0x75,
        0x08,  #  Report Size (8)                    25
        0x95,
        0x06,  #  Report Count (6)                   27
        0x81,
        0x02,  #  Input (Data,Var,Abs)               29
        0x06,
        0x00,
        0xFF,  #  Usage Page (Vendor Defined Page 1) 31
        0x09,
        0x20,  #  Usage (Vendor Usage 0x20)          34
        0x95,
        0x01,  #  Report Count (1)                   36
        0x81,
        0x02,  #  Input (Data,Var,Abs)               38
        0x05,
        0x01,  #  Usage Page (Generic Desktop)       40
        0x09,
        0x39,  #  Usage (Hat switch)                 42
        0x15,
        0x00,  #  Logical Minimum (0)                44
        0x25,
        0x07,  #  Logical Maximum (7)                46
        0x35,
        0x00,  #  Physical Minimum (0)               48
        0x46,
        0x3B,
        0x01,  #  Physical Maximum (315)             50
        0x65,
        0x14,  #  Unit (EnglishRotation: deg)        53
        0x75,
        0x04,  #  Report Size (4)                    55
        0x95,
        0x01,  #  Report Count (1)                   57
        0x81,
        0x42,  #  Input (Data,Var,Abs,Null)          59
        0x65,
        0x00,  #  Unit (None)                        61
        0x05,
        0x09,  #  Usage Page (Button)                63
        0x19,
        0x01,  #  Usage Minimum (1)                  65
        0x29,
        0x0F,  #  Usage Maximum (15)                 67
        0x15,
        0x00,  #  Logical Minimum (0)                69
        0x25,
        0x01,  #  Logical Maximum (1)                71
        0x75,
        0x01,  #  Report Size (1)                    73
        0x95,
        0x0F,  #  Report Count (15)                  75
        0x81,
        0x02,  #  Input (Data,Var,Abs)               77
        0x06,
        0x00,
        0xFF,  #  Usage Page (Vendor Defined Page 1) 79
        0x09,
        0x21,  #  Usage (Vendor Usage 0x21)          82
        0x95,
        0x0D,  #  Report Count (13)                  84
        0x81,
        0x02,  #  Input (Data,Var,Abs)               86
        0x06,
        0x00,
        0xFF,  #  Usage Page (Vendor Defined Page 1) 88
        0x09,
        0x22,  #  Usage (Vendor Usage 0x22)          91
        0x15,
        0x00,  #  Logical Minimum (0)                93
        0x26,
        0xFF,
        0x00,  #  Logical Maximum (255)              95
        0x75,
        0x08,  #  Report Size (8)                    98
        0x95,
        0x34,  #  Report Count (52)                  100
        0x81,
        0x02,  #  Input (Data,Var,Abs)               102
        0x85,
        0x02,  #  Report ID (2)                      104
        0x09,
        0x23,  #  Usage (Vendor Usage 0x23)          106
        0x95,
        0x3F,  #  Report Count (63)                  108
        0x91,
        0x02,  #  Output (Data,Var,Abs)              110
        0x85,
        0x05,  #  Report ID (5)                      112
        0x09,
        0x33,  #  Usage (Vendor Usage 0x33)          114
        0x95,
        0x28,  #  Report Count (40)                  116
        0xB1,
        0x02,  #  Feature (Data,Var,Abs)             118
        0x85,
        0x08,  #  Report ID (8)                      120
        0x09,
        0x34,  #  Usage (Vendor Usage 0x34)          122
        0x95,
        0x2F,  #  Report Count (47)                  124
        0xB1,
        0x02,  #  Feature (Data,Var,Abs)             126
        0x85,
        0x09,  #  Report ID (9)                      128
        0x09,
        0x24,  #  Usage (Vendor Usage 0x24)          130
        0x95,
        0x13,  #  Report Count (19)                  132
        0xB1,
        0x02,  #  Feature (Data,Var,Abs)             134
        0x85,
        0x0A,  #  Report ID (10)                     136
        0x09,
        0x25,  #  Usage (Vendor Usage 0x25)          138
        0x95,
        0x1A,  #  Report Count (26)                  140
        0xB1,
        0x02,  #  Feature (Data,Var,Abs)             142
        0x85,
        0x20,  #  Report ID (32)                     144
        0x09,
        0x26,  #  Usage (Vendor Usage 0x26)          146
        0x95,
        0x3F,  #  Report Count (63)                  148
        0xB1,
        0x02,  #  Feature (Data,Var,Abs)             150
        0x85,
        0x21,  #  Report ID (33)                     152
        0x09,
        0x27,  #  Usage (Vendor Usage 0x27)          154
        0x95,
        0x04,  #  Report Count (4)                   156
        0xB1,
        0x02,  #  Feature (Data,Var,Abs)             158
        0x85,
        0x22,  #  Report ID (34)                     160
        0x09,
        0x40,  #  Usage (Vendor Usage 0x40)          162
        0x95,
        0x3F,  #  Report Count (63)                  164
        0xB1,
        0x02,  #  Feature (Data,Var,Abs)             166
        0x85,
        0x80,  #  Report ID (128)                    168
        0x09,
        0x28,  #  Usage (Vendor Usage 0x28)          170
        0x95,
        0x3F,  #  Report Count (63)                  172
        0xB1,
        0x02,  #  Feature (Data,Var,Abs)             174
        0x85,
        0x81,  #  Report ID (129)                    176
        0x09,
        0x29,  #  Usage (Vendor Usage 0x29)          178
        0x95,
        0x3F,  #  Report Count (63)                  180
        0xB1,
        0x02,  #  Feature (Data,Var,Abs)             182
        0x85,
        0x82,  #  Report ID (130)                    184
        0x09,
        0x2A,  #  Usage (Vendor Usage 0x2a)          186
        0x95,
        0x09,  #  Report Count (9)                   188
        0xB1,
        0x02,  #  Feature (Data,Var,Abs)             190
        0x85,
        0x83,  #  Report ID (131)                    192
        0x09,
        0x2B,  #  Usage (Vendor Usage 0x2b)          194
        0x95,
        0x3F,  #  Report Count (63)                  196
        0xB1,
        0x02,  #  Feature (Data,Var,Abs)             198
        0x85,
        0x84,  #  Report ID (132)                    200
        0x09,
        0x2C,  #  Usage (Vendor Usage 0x2c)          202
        0x95,
        0x3F,  #  Report Count (63)                  204
        0xB1,
        0x02,  #  Feature (Data,Var,Abs)             206
        0x85,
        0x85,  #  Report ID (133)                    208
        0x09,
        0x2D,  #  Usage (Vendor Usage 0x2d)          210
        0x95,
        0x02,  #  Report Count (2)                   212
        0xB1,
        0x02,  #  Feature (Data,Var,Abs)             214
        0x85,
        0xA0,  #  Report ID (160)                    216
        0x09,
        0x2E,  #  Usage (Vendor Usage 0x2e)          218
        0x95,
        0x01,  #  Report Count (1)                   220
        0xB1,
        0x02,  #  Feature (Data,Var,Abs)             222
        0x85,
        0xE0,  #  Report ID (224)                    224
        0x09,
        0x2F,  #  Usage (Vendor Usage 0x2f)          226
        0x95,
        0x3F,  #  Report Count (63)                  228
        0xB1,
        0x02,  #  Feature (Data,Var,Abs)             230
        0x85,
        0xF0,  #  Report ID (240)                    232
        0x09,
        0x30,  #  Usage (Vendor Usage 0x30)          234
        0x95,
        0x3F,  #  Report Count (63)                  236
        0xB1,
        0x02,  #  Feature (Data,Var,Abs)             238
        0x85,
        0xF1,  #  Report ID (241)                    240
        0x09,
        0x31,  #  Usage (Vendor Usage 0x31)          242
        0x95,
        0x3F,  #  Report Count (63)                  244
        0xB1,
        0x02,  #  Feature (Data,Var,Abs)             246
        0x85,
        0xF2,  #  Report ID (242)                    248
        0x09,
        0x32,  #  Usage (Vendor Usage 0x32)          250
        0x95,
        0x34,  #  Report Count (52)                  252
        0xB1,
        0x02,  #  Feature (Data,Var,Abs)             254
        0x85,
        0xF4,  #  Report ID (244)                    256
        0x09,
        0x35,  #  Usage (Vendor Usage 0x35)          258
        0x95,
        0x3F,  #  Report Count (63)                  260
        0xB1,
        0x02,  #  Feature (Data,Var,Abs)             262
        0x85,
        0xF5,  #  Report ID (245)                    264
        0x09,
        0x36,  #  Usage (Vendor Usage 0x36)          266
        0x95,
        0x03,  #  Report Count (3)                   268
        0xB1,
        0x02,  #  Feature (Data,Var,Abs)             270
        0x85,
        0x60,  #  Report ID (96)                     272
        0x09,
        0x41,  #  Usage (Vendor Usage 0x41)          274
        0x95,
        0x3F,  #  Report Count (63)                  276
        0xB1,
        0x02,  #  Feature (Data,Var,Abs)             278
        0x85,
        0x61,  #  Report ID (97)                     280
        0x09,
        0x42,  #  Usage (Vendor Usage 0x42)          282
        0xB1,
        0x02,  #  Feature (Data,Var,Abs)             284
        0x85,
        0x62,  #  Report ID (98)                     286
        0x09,
        0x43,  #  Usage (Vendor Usage 0x43)          288
        0xB1,
        0x02,  #  Feature (Data,Var,Abs)             290
        0x85,
        0x63,  #  Report ID (99)                     292
        0x09,
        0x44,  #  Usage (Vendor Usage 0x44)          294
        0xB1,
        0x02,  #  Feature (Data,Var,Abs)             296
        0x85,
        0x64,  #  Report ID (100)                    298
        0x09,
        0x45,  #  Usage (Vendor Usage 0x45)          300
        0xB1,
        0x02,  #  Feature (Data,Var,Abs)             302
        0x85,
        0x65,  #  Report ID (101)                    304
        0x09,
        0x46,  #  Usage (Vendor Usage 0x46)          306
        0xB1,
        0x02,  #  Feature (Data,Var,Abs)             308
        0x85,
        0x68,  #  Report ID (104)                    310
        0x09,
        0x47,  #  Usage (Vendor Usage 0x47)          312
        0xB1,
        0x02,  #  Feature (Data,Var,Abs)             314
        0x85,
        0x70,  #  Report ID (112)                    316
        0x09,
        0x48,  #  Usage (Vendor Usage 0x48)          318
        0xB1,
        0x02,  #  Feature (Data,Var,Abs)             320
        0x85,
        0x71,  #  Report ID (113)                    322
        0x09,
        0x49,  #  Usage (Vendor Usage 0x49)          324
        0xB1,
        0x02,  #  Feature (Data,Var,Abs)             326
        0x85,
        0x72,  #  Report ID (114)                    328
        0x09,
        0x4A,  #  Usage (Vendor Usage 0x4a)          330
        0xB1,
        0x02,  #  Feature (Data,Var,Abs)             332
        0x85,
        0x73,  #  Report ID (115)                    334
        0x09,
        0x4B,  #  Usage (Vendor Usage 0x4b)          336
        0xB1,
        0x02,  #  Feature (Data,Var,Abs)             338
        0x85,
        0x74,  #  Report ID (116)                    340
        0x09,
        0x4C,  #  Usage (Vendor Usage 0x4c)          342
        0xB1,
        0x02,  #  Feature (Data,Var,Abs)             344
        0x85,
        0x75,  #  Report ID (117)                    346
        0x09,
        0x4D,  #  Usage (Vendor Usage 0x4d)          348
        0xB1,
        0x02,  #  Feature (Data,Var,Abs)             350
        0x85,
        0x76,  #  Report ID (118)                    352
        0x09,
        0x4E,  #  Usage (Vendor Usage 0x4e)          354
        0xB1,
        0x02,  #  Feature (Data,Var,Abs)             356
        0x85,
        0x77,  #  Report ID (119)                    358
        0x09,
        0x4F,  #  Usage (Vendor Usage 0x4f)          360
        0xB1,
        0x02,  #  Feature (Data,Var,Abs)             362
        0x85,
        0x78,  #  Report ID (120)                    364
        0x09,
        0x50,  #  Usage (Vendor Usage 0x50)          366
        0xB1,
        0x02,  #  Feature (Data,Var,Abs)             368
        0x85,
        0x79,  #  Report ID (121)                    370
        0x09,
        0x51,  #  Usage (Vendor Usage 0x51)          372
        0xB1,
        0x02,  #  Feature (Data,Var,Abs)             374
        0x85,
        0x7A,  #  Report ID (122)                    376
        0x09,
        0x52,  #  Usage (Vendor Usage 0x52)          378
        0xB1,
        0x02,  #  Feature (Data,Var,Abs)             380
        0x85,
        0x7B,  #  Report ID (123)                    382
        0x09,
        0x53,  #  Usage (Vendor Usage 0x53)          384
        0xB1,
        0x02,  #  Feature (Data,Var,Abs)             386
        0xC0,  # End Collection                      388
    ]
)

DS5_EDGE_MAC_ADDR = [0x74, 0xE7, 0xD6, 0x3A, 0x47, 0xE8]
DS5_EDGE_STOCK_REPORTS = {
    0x09: bytes(  # Pairing
        [
            0x09,
            DS5_EDGE_MAC_ADDR[0],
            DS5_EDGE_MAC_ADDR[1],
            DS5_EDGE_MAC_ADDR[2],
            DS5_EDGE_MAC_ADDR[3],
            DS5_EDGE_MAC_ADDR[4],
            DS5_EDGE_MAC_ADDR[5],
            0x08,
            0x25,
            0x00,
            0x1E,
            0x00,
            0xEE,
            0x74,
            0xD0,
            0xBC,
            0x00,
            0x00,
            0x00,
            0x00,
        ]
    ),
    0x20: bytes(  # Firmware Info
        [
            0x20,
            0x4A,
            0x75,
            0x6E,
            0x20,
            0x31,
            0x39,
            0x20,
            0x32,
            0x30,
            0x32,
            0x33,
            0x31,
            0x34,
            0x3A,
            0x34,
            0x37,
            0x3A,
            0x33,
            0x34,
            0x03,
            0x00,
            0x44,
            0x00,
            0x08,
            0x02,
            0x00,
            0x01,
            0x36,
            0x00,
            0x00,
            0x01,
            0xC1,
            0xC8,
            0x00,
            0x00,
            0x00,
            0x00,
            0x00,
            0x00,
            0x00,
            0x00,
            0x00,
            0x00,
            0x54,
            0x01,
            0x00,
            0x00,
            0x14,
            0x00,
            0x00,
            0x00,
            0x0B,
            0x00,
            0x01,
            0x00,
            0x06,
            0x00,
            0x00,
            0x00,
            0x00,
            0x00,
            0x00,
            0x00,
        ]
    ),
    0x05: bytes(  # Calibration
        [
            0x05,
            0xFE,
            0xFF,
            0xFC,
            0xFF,
            0xFE,
            0xFF,
            0x83,
            0x22,
            0x78,
            0xDD,
            0x92,
            0x22,
            0x5F,
            0xDD,
            0x95,
            0x22,
            0x6D,
            0xDD,
            0x1C,
            0x02,
            0x1C,
            0x02,
            0xF2,
            0x1F,
            0xED,
            0xDF,
            0xE3,
            0x20,
            0xDA,
            0xE0,
            0xEE,
            0x1F,
            0xDF,
            0xDF,
            0x0B,
            0x00,
            0x00,
            0x00,
            0x00,
            0x00,
        ]
    ),
}



import select
from time import sleep
from typing import Any, Generator, Literal, NamedTuple, Sequence

from anyio import open_process

from hhd.controller.base import Event
from ..base import Axis, Producer
import os

import logging

logger = logging.getLogger(__name__)


class ScanElement(NamedTuple):
    # Output Info
    axis: Axis
    # Buffer Info
    endianness: Literal["little", "big"]
    signed: bool
    bits: int
    storage_bits: int
    shift: int
    # Postprocess info
    scale: float
    offset: float


class DeviceInfo(NamedTuple):
    dev: str
    axis: Sequence[ScanElement]


ACCEL_MAPPINGS = [
    ("accel_z", "accel_x", "accel"),
    ("accel_x", "accel_y", "accel"),
    ("accel_y", "accel_z", "accel"),
]
GYRO_MAPPINGS = [
    ("gyro_z", "anglvel_x", "anglvel"),
    ("gyro_x", "anglvel_y", "anglvel"),
    ("gyro_y", "anglvel_z", "anglvel"),
]


def find_sensor(sensor: str):
    IIO_BASE_DIR = "/sys/bus/iio/devices/"

    for d in os.listdir(IIO_BASE_DIR):
        if not "device" in d:
            continue

        sensor_dir = os.path.join(IIO_BASE_DIR, d)
        name_fn = os.path.join(IIO_BASE_DIR, d, "name")

        if not os.path.isfile(name_fn):
            continue

        with open(name_fn, "r") as f:
            name = f.read()

        if name.strip() == sensor:
            logger.info(f"Found device '{sensor}' at\n{sensor_dir}")
            return sensor_dir

    return None


def write_sysfs(dir: str, fn: str, val: Any):
    with open(os.path.join(dir, fn), "w") as f:
        f.write(str(val))


def read_sysfs(dir: str, fn: str):
    with open(os.path.join(dir, fn), "r") as f:
        return f.read().strip()


def prepare_dev(
    sensor_dir: str, mappings: Sequence[tuple[Axis, str, str]]
) -> DeviceInfo | None:
    # @TODO: Add boosting sampling frequency support

    # Prepare device buffer
    dev = os.path.join("/dev", os.path.basename(sensor_dir))
    axis = {}
    write_sysfs(sensor_dir, "buffer/enable", 0)

    # Disable all scan elements
    for s in os.listdir(os.path.join(sensor_dir, "scan_elements")):
        if s.endswith("_en"):
            write_sysfs(sensor_dir, os.path.join("scan_elements", s), 0)

    # Selectively enable required ones and fill up buffer
    for ax, fn, meta_fn in mappings:
        write_sysfs(sensor_dir, f"scan_elements/in_{fn}_en", 1)

        # Prepare buffer metadata
        idx = int(read_sysfs(sensor_dir, f"scan_elements/in_{fn}_index"))
        if idx == -1:
            logger.error(
                f"Device '{dev}' element '{fn}' does not support buffered capture."
            )
            return None
        se = read_sysfs(sensor_dir, f"scan_elements/in_{fn}_type")

        endianness = "big" if se.startswith("be:") else "little"
        signed = "e:s" in se

        bits = int(se[se.index(":") + 2 : se.index(":") + 4])
        storage_bits = int(se[se.index("/") + 1 : se.index("/") + 3])
        shift = int(se[-1])

        # Prepare scan metadata
        offset = float(read_sysfs(sensor_dir, f"in_{meta_fn}_offset"))
        scale = float(read_sysfs(sensor_dir, f"in_{meta_fn}_scale"))

        axis[idx] = ScanElement(
            ax, endianness, signed, bits, storage_bits, shift, scale, offset
        )
    write_sysfs(sensor_dir, "buffer/enable", 1)

    axis_arr = tuple(axis[i] for i in sorted(axis))
    return DeviceInfo(dev, axis_arr)


def process_scan_event(data: bytes, ofs: int, se: ScanElement):
    # TODO: Implement parsing iio fully, by adding shifting and cutoff
    d = data[ofs >> 3 : (ofs >> 3) + (se.storage_bits >> 3)]
    d = int.from_bytes(d, byteorder=se.endianness, signed=se.signed)
    # d = d >> se.shift
    # d &= (1 << se.bits) - 1
    d = d * se.scale + se.offset
    return d


def get_size(dev: DeviceInfo):
    out = 0
    for s in dev.axis:
        out += s.storage_bits
    return out >> 3


class Imu(Producer):
    def __init__(self, name: str, mappings) -> None:
        self.name = name
        self.mappings = mappings
        self.fd = 0

    def open(self):
        sens_dir = find_sensor(self.name)
        if not sens_dir:
            return []

        dev = prepare_dev(sens_dir, self.mappings)
        if not dev:
            return []

        self.dev = dev
        self.fd = os.open(dev.dev, os.O_RDONLY)
        self.size = get_size(dev)

        return [self.fd]

    def close(self, exit: bool):
        if self.fd:
            os.close(self.fd)
        self.fd = 0
        return True

    def produce(self, fds: Sequence[int]) -> Sequence[Event]:
        if self.fd not in fds:
            return []

        data = os.read(self.fd, self.size)

        # Empty the buffer preventing repeated calls
        while select.select([self.fd], [], [], 0)[0]:
            data = os.read(self.fd, self.size)

        out = []
        ofs = 0
        for se in self.dev.axis:
            out.append(
                {
                    "type": "axis",
                    "axis": se.axis,
                    "val": process_scan_event(data, ofs, se),
                }
            )
            ofs += se.storage_bits
        return out


class AccelImu(Imu):
    def __init__(self) -> None:
        super().__init__("accel_3d", ACCEL_MAPPINGS)


class GyroImu(Imu):
    def __init__(self) -> None:
        super().__init__("gyro_3d", GYRO_MAPPINGS)



import sys
from threading import Lock, Condition
import time
from typing import Sequence

from hhd.controller.base import Axis, Button, Event

from ..base import Consumer, Producer
from ..uhid import UhidDevice
from .const import (
    DS5_EDGE_BUS,
    DS5_EDGE_COUNTRY,
    DS5_EDGE_DESCRIPTOR,
    DS5_EDGE_MAX_REPORT_FREQ,
    DS5_EDGE_MIN_REPORT_FREQ,
    DS5_EDGE_NAME,
    DS5_EDGE_PRODUCT,
    DS5_EDGE_REPORT_USB_BASE,
    DS5_EDGE_STOCK_REPORTS,
    DS5_EDGE_VENDOR,
    DS5_EDGE_VERSION,
    DS5_EDGE_DELTA_TIME,
)

REPORT_MAX_DELAY = 1 / DS5_EDGE_MIN_REPORT_FREQ
REPORT_MIN_DELAY = 1 / DS5_EDGE_MAX_REPORT_FREQ
ANGL_RES = 1024


class DualSense5Edge(Producer, Consumer):
    def __init__(self) -> None:
        self.available = False
        self.report = None
        self.dev = None
        self.start = 0

    def open(self) -> Sequence[int]:
        self.available = False
        self.report = bytearray(DS5_EDGE_REPORT_USB_BASE)
        self.dev = UhidDevice(
            vid=DS5_EDGE_VENDOR,
            pid=DS5_EDGE_PRODUCT,
            bus=DS5_EDGE_BUS,
            version=DS5_EDGE_VERSION,
            country=DS5_EDGE_COUNTRY,
            name=DS5_EDGE_NAME,
            report_descriptor=DS5_EDGE_DESCRIPTOR,
        )
        self.start = time.time()
        self.fd = self.dev.open()
        return [self.fd]

    def close(self, exit: bool) -> bool:
        if not exit:
            """This is a consumer, so we would deadlock if it was disabled."""
            return False

        if self.dev:
            self.dev.send_destroy()
            self.dev.close()
            self.dev = None
            self.fd = None

        return True

    def produce(self, fds: Sequence[int]) -> Sequence[Event]:
        if self.fd not in fds:
            return []

        # Process queued events
        assert self.dev
        while ev := self.dev.read_event():
            match ev["type"]:
                case "open":
                    self.available = True
                case "close":
                    self.available = False
                case "get_report":
                    if ev["rnum"] in DS5_EDGE_STOCK_REPORTS:
                        self.dev.send_get_report_reply(
                            ev["id"], 0, DS5_EDGE_STOCK_REPORTS[ev["rnum"]]
                        )

        return []

    def consume(self, events: Sequence[Event]):
        assert self.dev and self.report
        for ev in events:
            match ev["type"]:
                case "axis":
                    type = None
                    ofs = None
                    match ev["axis"]:
                        case "gyro_x":
                            ofs = 16
                            type = "gyro"
                        case "gyro_y":
                            ofs = 18
                            type = "gyro"
                        case "gyro_z":
                            ofs = 20
                            type = "gyro"
                        case "accel_x":
                            ofs = 22
                            type = "accel"
                        case "accel_y":
                            ofs = 24
                            type = "accel"
                        case "accel_z":
                            ofs = 26
                            type = "accel"

                    if not type or not ofs:
                        continue

                    val = ev["val"]
                    # TODO: Figure out the correct normalization values
                    # For now, this does the inverse scaling of the legion go's imu data
                    if type == "gyro":
                        val = 5729.6 * val
                    elif type == "accel":
                        val = 10.19716 * val
                    val = int(val)

                    try:
                        self.report[ofs : ofs + 2] = int.to_bytes(
                            val, length=2, byteorder=sys.byteorder, signed=True
                        )
                    except:
                        # TODO: Debug
                        pass

        # Send report
        self.report[28:32] = int(
            (time.time() - self.start) * DS5_EDGE_DELTA_TIME
        ).to_bytes(4, byteorder=sys.byteorder, signed=False)
        self.dev.send_input_report(self.report)



from typing import Any, Literal, Sequence, TypedDict

Axis = Literal[
    # Sticks
    # Values should range from -1 to 1
    "ls_x",
    "ls_y",
    "rs_x",
    "rs_y",
    # Triggers
    # Values should range from -1 to 1
    "lt",
    "rt",
    # Hat, implemented as axis. Either -1, 0, or 1
    "hat_x",
    "hat_y",
    # Accelerometer
    # Values should be in m2/s
    "accel_x",
    "accel_y",
    "accel_z",
    # Gyroscope
    # Values should be in deg/s
    "gyro_x",
    "gyro_y",
    "gyro_z",
    # Touchpad
    # Both width and height should go from [0, 1]. Aspect ratio is a setting.
    # It is up to the device whether to stretch, crop and how to crop (either 
    # crop part of the input or part of itself)
    "touchpad_x",
    "touchpad_y",
]


Button = Literal[
    # Thumbpad
    "a",
    "b",
    "x",
    "y",
    # Sticks
    "ls",
    "rs",
    # Bumpers
    "lb",
    "rb",
    # Back buttons
    "extra_l1",
    "extra_l2",
    "extra_l3",
    "extra_r1",
    "extra_r2",
    "extra_r3",
    # Select
    "start",
    "select",
    # Misc
    "mode",
    "share",
    "touchpad_touch",
    "touchpad_click",
]


Configuration = Literal[
    # Misc
    "led_mute",  # binary
    "player",
    # Set the aspect ratio of the touchpad used
    "touchpad_aspect_ratio",
]


class RumbleEvent(TypedDict):
    type: Literal["rumble"]
    # Rumble side, if the controller has left/right rubmles.
    # Producers that control rumble should either use `left`/`right` or `both`.
    # But not both. Consumers that support 2 rumble and see `both` should set the
    # color of both rumble. Consumers that have a single rumble and see 'left` or `right`
    # shall use the last value or pick a random side (TODO: Determine if there
    # are issues with either approach).
    side: Literal["both", "left", "right"]
    mode: Literal["disable", "square", "cross", "circle", "triangle"]
    intensity: float


class LedEvent(TypedDict):
    type: Literal["led"]

    # Led side, if the controller has left/right leds.
    # Producers that control leds should either use `left`/`right` or `both`.
    # But not both. Consumers that support 2 leds and see `both` should set the
    # color of both leds. Consumers that support 1 LED and see 'left` or `right`
    # shall use the last value or pick a random side (TODO: Determine if there
    # are issues with either approach).
    led: Literal["both", "left", "right"]

    # Various lighting modes supported by the led.
    mode: Literal["disable", "solid", "blinking", "rainbow", "spiral"]

    # Brightness range is from 0 to 1
    # If the response device does not support brightness control, it shall
    # convert the rgb color to hue, assume saturation is 1, and derive a new
    # RGB value from the brightness below
    brightness: float

    # The speed the led should blink if supported by the led
    speed: float

    # Color values for the led, may be ommited depending on the mode, by being
    # set to 0
    red: int
    green: int
    blue: int


class ButtonEvent(TypedDict):
    type: Literal["button"]
    button: Button
    held: int


class AxisEvent(TypedDict):
    type: Literal["axis"]
    axis: Axis
    val: float


class ConfigurationEvent(TypedDict):
    type: Literal["configuration"]
    conf: Configuration
    val: Any


Event = RumbleEvent | ButtonEvent | AxisEvent | ConfigurationEvent


class Producer:
    def open(self) -> Sequence[int]:
        """Opens and returns a list of file descriptors that should be listened to."""
        raise NotImplementedError()

    def close(self, exit: bool) -> bool:
        """Called to close the device.

        If `exit` is true, the program is about to
        close. If it is false, the controller is entering power save mode because
        it is unused. In this case, if this service is required, you may forgo
        closing and return false. If true, it is assumed this producer is closed.

        `open()` will be called again once the consumers are ready."""
        return False

    def produce(self, fds: Sequence[int]) -> Sequence[Event]:
        """Called with the file descriptors that are ready to read."""
        raise NotImplementedError()


class Consumer:
    available: bool
    """Hint that states if the consumer can receive events. If it is false,
    consumer will not be called. If all consumers are false, producers will
    be closed to save CPU utilisation."""

    def initialize(self):
        """Optional method for initialization."""
        pass

    def consume(self, events: Sequence[Event]):
        pass



import select
from typing import Sequence, TypeVar, cast

import evdev

from hhd.controller.base import Event

from ..base import Axis, Button, Producer

import logging

logger = logging.getLogger(__name__)


def B(b: str):
    return cast(int, getattr(evdev.ecodes, b))


A = TypeVar("A")


def to_map(b: dict[A, Sequence[int]]) -> dict[int, A]:
    out = {}
    for btn, seq in b.items():
        for s in seq:
            out[s] = btn
    return out


XBOX_BUTTON_MAP: dict[int, Button] = to_map(
    {
        # Gamepad
        "a": [B("BTN_A")],
        "b": [B("BTN_B")],
        "x": [B("BTN_X")],
        "y": [B("BTN_Y")],
        # Sticks
        "ls": [B("BTN_THUMBL")],
        "rs": [B("BTN_THUMBR")],
        # Bumpers
        "lb": [B("BTN_TL")],
        "rb": [B("BTN_TR")],
        # Select
        "start": [B("BTN_START")],
        "select": [B("BTN_SELECT")],
        # Misc
        "mode": [B("BTN_MODE")],
    }
)

XBOX_AXIS_MAP: dict[int, Axis] = to_map(
    {
        # Sticks
        # Values should range from -1 to 1
        "ls_x": [B("ABS_X")],
        "ls_y": [B("ABS_Y")],
        "rs_x": [B("ABS_RX")],
        "rs_y": [B("ABS_RY")],
        # Triggers
        # Values should range from -1 to 1
        "rt": [B("ABS_Z")],
        "lt": [B("ABS_RZ")],
        # Hat, implemented as axis. Either -1, 0, or 1
        "hat_x": [B("ABS_HAT0X")],
        "hat_y": [B("ABS_HAT0Y")],
    }
)

LGO_TOUCHPAD_BUTTON_MAP: dict[int, Button] = to_map(
    {
        "touchpad_touch": [B("BTN_TOOL_FINGER")],  # also BTN_TOUCH
        "touchpad_click": [B("BTN_TOOL_DOUBLETAP")],
    }
)

LGO_TOUCHPAD_AXIS_MAP: dict[int, Axis] = to_map(
    {
        "touchpad_x": [B("ABS_X")],  # also ABS_MT_POSITION_X
        "touchpad_y": [B("ABS_Y")],  # also ABS_MT_POSITION_Y
    }
)


class GenericGamepadEvdev(Producer):
    def __init__(
        self,
        vid: int | None,
        pid: int | None,
        name: str | None,
        btn_map: dict[int, Button] = XBOX_BUTTON_MAP,
        axis_map: dict[int, Axis] = XBOX_AXIS_MAP,
        aspect_ratio: float | None = None,
    ) -> None:
        self.vid = vid
        self.pid = pid
        self.name = name

        self.btn_map = btn_map
        self.axis_map = axis_map
        self.aspect_ratio = aspect_ratio

        self.dev: evdev.InputDevice | None = None
        self.fd = 0

    def open(self) -> Sequence[int]:
        for d in evdev.list_devices():
            dev = evdev.InputDevice(d)
            if self.vid and dev.info.vendor != self.vid:
                continue
            if self.pid and dev.info.product != self.pid:
                continue
            if self.name and dev.name != self.name:
                continue
            self.dev = dev
            self.dev.grab()
            self.ranges = {
                a: (i.min, i.max) for a, i in self.dev.capabilities()[B("EV_ABS")]  # type: ignore
            }
            self.fd = dev.fd
            self.started = True
            return [self.fd]

        logger.error(
            f"Device not found:\n{(self.pid if self.pid else 0):04X}:{(self.vid if self.vid else 0):04X} {self.name}"
        )
        return []

    def close(self, exit: bool) -> bool:
        if self.dev:
            self.dev.close()
            self.fd = 0
        return True

    def produce(self, fds: Sequence[int]) -> Sequence[Event]:
        if not self.dev or not self.fd in fds:
            return []

        out: list[Event] = []
        if self.started and self.aspect_ratio is not None:
            self.started = False
            out.append(
                {
                    "type": "configuration",
                    "conf": "touchpad_aspect_ratio",
                    "val": self.aspect_ratio,
                }
            )

        while select.select([self.fd], [], [], 0)[0]:
            for e in self.dev.read():
                if e.type == B("EV_KEY"):
                    if e.code in self.btn_map:
                        out.append(
                            {
                                "type": "button",
                                "button": self.btn_map[e.code],
                                "held": bool(e.value),
                            }
                        )
                elif e.type == B("EV_ABS"):
                    if e.code in self.axis_map:
                        # Normalize
                        val = e.value / abs(
                            self.ranges[e.code][1 if e.value >= 0 else 0]
                        )

                        out.append(
                            {
                                "type": "axis",
                                "axis": self.axis_map[e.code],
                                "val": val,
                            }
                        )
        return out
